/*
 * This is a netscape bug workaround based on the preload library
 * of Stanislav Meduna <stano@trillian.eunet.sk>.
 *
 * Changes for SuSE done by <werner@suse.de>
 *
 * * Don't fake the write of the MAGIC character anymore but
 *   simulate a interrupted `write' system call by setting
 *   errno to EINTR with written bytes of `-1'.
 *
 * * If a real `write' system call returns with the errno
 *   EAGAIN we map this errno to EINTR.  This because it seems
 *   that EAGAIN causes Netscape to listen on socket or pipe
 *   events which sometimes never happen.
 *
 * * If MALLOC_CHECK_ environment variable isn't set we wrap
 *   malloc functions by using the hooks of the glibc. With
 *   these hooks we try to make free more safety against
 *   doubled freed pointers.
 *
 * * Compile and link with
 *
 *   gcc -O2 -fPIC -nostartfiles -nostdlib -shared \
 *       -o libnspipepatch.so nspipepatch.c -lc -ldl
 *
 * * Install libnspipepatch.so to an appropiate path and call
 *   Netscape with the LD_PRELOAD environment variable set.
 */

#include <features.h>
#if defined(__GLIBC__) && defined(__GLIBC_MINOR__)
# define GLIBC_VERSION ((__GLIBC__ << 8) + __GLIBC_MINOR__)
#endif
#include <dlfcn.h>
#include <sys/types.h>
#include <limits.h>
#include <stdlib.h>
#include <errno.h>
#ifndef  USE_MALLOC_HOOKS
# define USE_MALLOC_HOOKS 1
# include <malloc.h>
#endif
#ifndef  DEFAULT_LIB
#define  DEFAULT_LIB "/lib/libc.so.6"
#endif
#ifndef  GETENV_OF
# define GETENV_OF 0
#endif

/* The magic netscape char flowing through the pipe */
#define MAGIC ((unsigned char) '\372')

/* libc handle for dlsym */
static void *dlhdl=0;

/* pointers to overloaded functions */
static int     (*pipe_fp) (int [2])=0;
static ssize_t (*write_fp)(int, const void *, size_t)=0;
static ssize_t (*read_fp) (int, void *, size_t)=0;
static int     (*close_fp)(int)=0;
static int     (*open_fp)(const char *, int, ...)=0;
#if GETENV_OF != 0
/*
 * Just for debugging: get all environment variables
 * used by Netscape and write it to /tmp/logg
 */
static char *  (*getenv_fp) (const char *)=0;
#endif

/* we keep the info about our descriptors here */
static struct nshack_pipe_t
{
  /* If this is a pipe that could be our one,
     mark both ends here. If we shouldn't touch
     the descriptor, enter -1 here
  */
  short read_end, write_end;

  /* The difference between writes and reads,
     maintained on the write end
  */
  int diff;

} fdflags[256];


/* Mark the descriptors as unused or not of our interest */
static inline void clearfd(int fd)
{
  fdflags[fd].read_end = fdflags[fd].write_end = -1;
  fdflags[fd].diff = 0;
}

/* Clear both ends */
static inline void clearboth(int fd)
{
  register int rd = fdflags[fd].read_end;
  register int wr = fdflags[fd].write_end;

  if (rd >= 0)
    clearfd(rd);
  if (wr >= 0)
    clearfd(wr);
}

#define FPsym(f,h)			\
if (!f##_fp) f##_fp = dlsym(h, #f);	\
if (!f##_fp) goto err;

#define OKsym(f)			\
if (!f##_fp) initptrs()

/* Init the pointers to overloaded functions */
static void initptrs()
{

  if (!dlhdl) {
    int i;
    char *lib = getenv("NSFIX_LIB");
    if (!lib)
	lib = DEFAULT_LIB;
    dlhdl = dlopen(lib, RTLD_NOW|RTLD_GLOBAL);
    for (i=0; i < sizeof(fdflags)/sizeof(fdflags[0]); i++)
      clearfd(i);
  }

  if (!dlhdl)
    goto err;

  FPsym(pipe, dlhdl);
  FPsym(write, dlhdl);
  FPsym(read, dlhdl);
  FPsym(close, dlhdl);
#if GETENV_OF != 0
  FPsym(getenv, dlhdl);
#endif

  if (dlhdl)
    dlclose(dlhdl);

  return;

err:
  exit(99);
}

/* Overloaded pipe - mark the fresh pipes */
int pipe(int fd[2])
{
  register int res;
  OKsym(pipe);

  errno = 0;
  res = (*pipe_fp)(fd);

  if (!res)
  {
    /* This could be our pipe - watch it during read/write */
    fdflags[fd[0]].read_end  = fdflags[fd[1]].read_end  = fd[0];
    fdflags[fd[0]].write_end = fdflags[fd[1]].write_end = fd[1];
    fdflags[fd[0]].diff      = fdflags[fd[1]].diff      = 0;
  }

  return res;
}


#include <stdio.h>
/* Overloaded write */
ssize_t write(int fd, const void *buf, size_t cnt)
{
  register int res=1;
  OKsym(write);

  /*
   * We use the netscape write handler to avoid pipe buffer overflow
   * by forcing the EINTR safe handling of that write handler.
   */
  if (fd == fdflags[fd].write_end && fdflags[fd].diff > _POSIX_PIPE_BUF)
  {
    res = -1;
    errno = EINTR;
    goto out;
  }

  errno = 0;
  res = (*write_fp)(fd, buf, cnt);

  /*
   * It seems that EAGAIN misleads netscape write handler to wait on
   * sockets/pipes events which do not happen anymore. Therefore we
   * use the EINTR safe handling done by the most programmers.
   */
  switch (errno) {
    case EAGAIN:
      errno = EINTR; res = -1;
      break;
    default:
      break;
  }

  if (fd == fdflags[fd].write_end && res > -1)
  {
    fdflags[fd].diff += res;
  }

out:
  return res;
}

/* Overloaded read */
ssize_t read(int fd, void *buf, size_t cnt)
{
  register int res=1;
  OKsym(read);

  errno = 0;
  res = (*read_fp)(fd, buf, cnt);

  if (fd == fdflags[fd].read_end && res > -1)
  {
    fdflags[fdflags[fd].write_end].diff -= res;
  }

  return res;
}

/* Overloaded close (we want to be at least a bit correct) */
int close(int fd)
{
  register int res=0;
  OKsym(close);

  if (fd == fdflags[fd].read_end || fd == fdflags[fd].write_end)
    clearboth(fd);
  res = (*close_fp)(fd);

  return res;
}

#if USE_MALLOC_HOOKS != 0
# if GLIBC_VERSION >= 0x201
static void * (*__malloc_hook_fp) (size_t, const void*)=0;
static void   (*__free_hook_fp) (void *, const void*)=0;
static void * (*__realloc_hook_fp) (void *, size_t, const void*)=0;
# else
static void * (*__malloc_hook_fp) (size_t)=0;
static void   (*__free_hook_fp) (void *)=0;
static void * (*__realloc_hook_fp) (void *, size_t)=0;
# endif

# if GLIBC_VERSION >= 0x201
static void * ns_malloc_hook (size_t size, const void *caller)
# else
static void * ns_malloc_hook (size_t size)
# endif
{
  register void *result = (void *)0;
# ifdef DO_ALIGN_ALLOC
  register size_t align = (size_t)((size + sizeof(long) - 1) & ~(sizeof(long) - 1));
# else
  register size_t align = size;
# endif

  __malloc_hook    =  __malloc_hook_fp;
  result = malloc (align);
  if (result)
    memset(result, 0, align);
  __malloc_hook_fp =  __malloc_hook;
  __malloc_hook    = ns_malloc_hook;

  return result;
}

# if GLIBC_VERSION >= 0x201
static void ns_free_hook (void *ptr, const void *caller)
# else
static void ns_free_hook (void *ptr)
# endif
{
  __free_hook    =  __free_hook_fp;
  if (ptr)
    free (ptr);
  ptr = (void *)0;
  __free_hook_fp =  __free_hook;
  __free_hook    = ns_free_hook;
}

# if GLIBC_VERSION >= 0x201
static void * ns_realloc_hook (void *ptr, size_t size, const void *caller)
# else
static void * ns_realloc_hook (void *ptr, size_t size)
# endif
{
  register void *result = (void *)0;

  __realloc_hook    =  __realloc_hook_fp;
  result = realloc(ptr, size);
  __realloc_hook_fp =  __realloc_hook;
  __realloc_hook    = ns_realloc_hook;

  return result;
}
#endif

#if GETENV_OF != 0
static int logg = -1;
# include <fcntl.h>
#endif
void _init()
{
#if USE_MALLOC_HOOKS != 0
  if (!getenv("MALLOC_CHECK_")) {
    __malloc_hook_fp  =   __malloc_hook;
    __malloc_hook     =  ns_malloc_hook;
    __free_hook_fp    =     __free_hook;
    __free_hook       =    ns_free_hook;
    __realloc_hook_fp =  __realloc_hook;
    __realloc_hook    = ns_realloc_hook;
  }
#endif

#if GETENV_OF != 0
  logg = open("/tmp/logg", O_WRONLY|O_CREAT|O_EXCL);
#endif
}

#if GETENV_OF != 0
char *getenv(const char *name)
{
  register char* value = (char*)0;
  OKsym(getenv);

  value = (*getenv_fp)(name);

  if (logg > 0) {
    write(logg, name,  strlen(name)* sizeof(char));
    if (value) {
      write(logg, "=", sizeof(char));
      write(logg, value, strlen(value)*sizeof(char));
    }
    write(logg, "\n", sizeof(char));
    fsync(logg);
  }

  return (*getenv_fp)(name);
}
#endif
